#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <stdlib.h>
#include <stdint.h>
#include <avr/pgmspace.h>

#include "main.h"

#include "driver/rs232.h"
//#include "driver/manchester.h"

volatile uint8_t datalen = 0;
char data[10]; // 8 bytes data buffer + string termination

static char buffer[RS232_BUFSIZE+1];
static uint8_t buffer_len;

volatile uint8_t bitbuf_len = 0;
volatile uint16_t bitbuf = 0;
volatile uint16_t bitbuf_out = 0;
volatile uint8_t bitbuf_out_len = 0;
volatile uint8_t timeout = 0;
volatile uint8_t lastbit = 0;



// USART0 RX interrupt
ISR ( USART_RXC_vect ) {
    char c = UDR;

    // check for buffer overflow
    if (buffer_len==sizeof(buffer)) {
        buffer_len=0;
        if (c == 27) {
            // escape sequence, store to empty buffer
            buffer[buffer_len++] = c;
        }
    }  else {
        // collect characters until end of line
        if (c == 27) {
            // escape sequence, clear buffer
            buffer_len = 0;
            buffer[buffer_len++] = c;
        } else if ( (c==0xff) && (buffer_len > 3) ) {
            buffer[buffer_len]=0;

            // packet end received, parse the received packet

            // wait for the next packet
            buffer_len=0;
        } else {
            buffer[buffer_len++]=c;
        }
    }
}


#define PULSE_IN        PIND
#define PULSE_BIT       2

#define TIMER_DIVISOR   8
#define VALUE_100US     0.47e-4
#define TIMER_100US     0xff - (uint8_t)(VALUE_100US * F_CPU/TIMER_DIVISOR)

ISR ( INT0_vect ) {
    writeBit(GICR, INT0, 0); // disable INT0 interrupt
    // Startsignal erkannt, ab hier den Timer0 starten,
    // der liest dann alle 100µs den Zustand ein und schreibt das
    // empfangene Bit in den Puffer
    lastbit = 0b10000001; // initialize

    TCNT0 = TIMER_100US;
    TIMSK |= 1<<TOIE0; //enable timer0 interrupt
}

ISR ( TIMER0_OVF_vect ) {
    uint8_t sr = SREG;
    TCNT0 = TIMER_100US; // reset timer for next bit time
    writeBit(PORTD, 4, 0);
    // read the current bit value, store it to 16bit buffer
    if ( (PULSE_IN & _BV(PULSE_BIT)) == 0 ) {
//RS232_putc('L');

        if ( (lastbit & 1) == 1) {
            // bit cleared, second block
            // insert bit
            bitbuf |= 1;
            bitbuf = bitbuf << 1;
            bitbuf_len++;
            lastbit = 0; // set block number0 & signal state to 0
        } else {
            // bit cleared, first block
            lastbit = 1; // set block number1 & signal state to 0
        }
    } else {
//RS232_putc('H');
        if ( (lastbit & 1) == 1) {
            // bit set, second block
            if ( (lastbit & _BV(7)) == 0) {
                // insert zero bit
                bitbuf = bitbuf << 1;
                bitbuf_len++;
                lastbit = 0b10000000; // set block number0 & signal state to 1
            } else {
                // no change, transmission END!
                lastbit = 0xff;
            }
        } else {
            // bit set, first block
            lastbit = 0b10000001; // set block number1 & signal state to 1
        }
    }
    if ( (bitbuf_len == 16) | (lastbit == 0xFF) ) {
        // 16 bits full or transmission end
        // export to uart
        writeBit(TIMSK, TOIE0, 0); // disable timer0 interrupt
        if (bitbuf > 0) {
            bitbuf_out = bitbuf;
            bitbuf_out_len = bitbuf_len;
        }
        bitbuf = 0;
        bitbuf_len = 0;
        GICR |= _BV(INT0); // Enable INT0
    }

    writeBit(PORTD, 4, 1);
    SREG = sr;
}

volatile uint16_t pulse_counter = 0;
volatile uint8_t measure_progress = 0;
volatile uint16_t car_id = 0;
ISR(TIMER2_OVF_vect) {
    TCNT2 = 0xF5;
    pulse_counter++;
}

ISR (INT1_vect) {
    // measure pulse width
    //RS232_putc('x');
    if (measure_progress == 0) {
        // start measuring
        TCNT2 = 0xF5;
        pulse_counter = 0;
        measure_progress = 1;
        // set INT1 to rising edge (stops current measurement)
        //MCUCR |= _BV(ISC10);
        writeBit(PORTD, 4, 1);
    } else {
        // stop measuring
        car_id = pulse_counter;
        measure_progress = 0;
        // set INT1 to falling edge (initiates next measurement)
        //MCUCR &= ~_BV(ISC10);
        writeBit(PORTD, 4, 0);
    }
}


int main(void)
{
  uint i;
char s[30];

    TCCR0 = 0;			//timer off
    //TCCR0 = (1<<CS00);			//divide by 1
    TCCR0 = (1<<CS01);			//divide by 8
    //TCCR0 = (1<<CS01) | (1<<CS00);			//divide by 64
    //TCCR0 = (1<<CS02);			//divide by 256
    //TCCR0 = (1<<CS02) | (1<<CS00);			//divide by 1024

    //TIMSK = 1<<TOIE0;			//enable timer interrupt
    // Timer will be enabled on INT0

    // configure timer2 for pulse measurement
    //TCCR2 = (1<<CS21); // divide 8
    TCCR2 = (1<<CS20); // divide 1
    TIMSK = 1<<TOIE2;

    // configure INT0+1 for falling edge
    // INT2 on every edge
    //MCUCR = _BV(ISC00)  | _BV(ISC11);
    MCUCR = _BV(ISC11);
//    GICR = _BV(INT0) | _BV(INT1) ; // Enable INT0/1
    GICR = _BV(INT1) ; // Enable INT1

    writeBit(DDRD, 4, 1);


    RS232_init(); // initialize RS232 interface
    RS232_puts_p(PSTR("INIT OK\n"));

    i = 0;
    
    sei();
    while (1) {
        // main loop

        if (car_id != 0) {
          itoa( car_id, s, 10);
          car_id = 0;
          RS232_puts( s );
          RS232_putc('\n');
            
        }
    /*cli();
    i = rc5_data; // read two bytes from interrupt !
    rc5_data = 0;
    sei();
    if (i) {
      itoa( i, s, 10);
      RS232_puts( s );
    } */
/*
        if (p_len < 0xFF) p_len++;
        if( (p_bit ^ PULSE_IN) & 1<<PULSE_BIT ){		// change detect
            p_bit = ~p_bit;				// 0x00 -> 0xFF -> 0x00
            itoa( p_len, s, 10);
            RS232_puts(s);
            RS232_puts_p(PSTR("\n"));
            p_len = 0;
        }
*/
    } // main loop end
};

